#include "arrays.h"
#include "NestStep.h"
#include "Combine.h"
#include <stdlib.h>
#include <mpi.h>
#include <string.h>
#include <stdio.h>

/*
 * new_RepArray
 * creates a new replicated array element
 */
RepArray * new_RepArray(Name _name, int _type, int _size, void * _base)
{
	RepArray * temparray;

	temparray = (RepArray *)malloc(sizeof(RepArray));
	temparray->name.procedure = _name.procedure;
	temparray->name.relative = _name.relative;
	temparray->type = _type;
	temparray->size = _size;
	temparray->base = _base;
	memcpy(&(temparray->name.path[0]), &(thisgroup->path[0]), MAX_DEPTH * sizeof(int));
	

	if(_type == IVAR)
	{
		temparray->hidden = (void *)malloc(_size * sizeof(int));
		memcpy(temparray->hidden, _base, _size * sizeof(int));
	}
	else if(_type == FVAR)
	{
		temparray->hidden = (void *)malloc(_size * sizeof(float));
		memcpy(temparray->hidden, _base, _size * sizeof(float));
	}
	else if(_type == DVAR)
	{
		temparray->hidden = (void *)malloc(_size * sizeof(double));
		memcpy(temparray->hidden, _base, _size * sizeof(double));
	}
	temparray->url = 0;
	temparray->urh = _size - 1;

	return temparray;
}

/*
 * free_RepArray
 * frees the memory occupied by the replicated array object
 * NOTE: does not free the memory for the actual array, only the
 * NestStep wrapper
 */
void free_RepArray(RepArray * _array)
{
	free(_array->hidden);
	free(_array);
}

// NestStep_RepArray_range
// set the range for which elements to update
int NestStep_RepArray_range(RepArray * _array, int _lower, int _higher)
{
	_array->url = _lower;
	_array->urh = _higher;
	return 0;
}

/*
 * NestStep_RepArray_attach
 * adds a replicated array to be combined at the end of the step
 */
int NestStep_RepArray_attach(RepArray * _array, 
							 int _combine_strategy,
							 void (*_combineFn)(void *, void *, void *),
							 int _prefix_tag,
							 void * _prefix_base,
							 int _updated, 
							 CombineList * _clist)
{
	int i;
	int size;
	CombineList * combineList;
	if(_clist == NULL)
		combineList = thisgroup->clist;
	else
		combineList = _clist;
	switch(_array->type) {
		case IVAR:
			size = _array->size * sizeof(int);
			break;
		case FVAR:
			size = _array->size * sizeof(float);
			break;
		case DVAR:
			size = _array->size * sizeof(double);
			break;
		default:
			return ATTACH_ERROR;
	}

	combineList->rep_info[combineList->no_rep_arrays].rep_array = _array;
	combineList->rep_info[combineList->no_rep_arrays].type = _array->type;
	combineList->rep_info[combineList->no_rep_arrays].combine_strategy = _combine_strategy;
	combineList->rep_info[combineList->no_rep_arrays].combineFn = _combineFn;
	combineList->rep_info[combineList->no_rep_arrays].prefix = _prefix_tag;
	combineList->rep_info[combineList->no_rep_arrays].prefix_base = _prefix_base;
	if(_prefix_tag == PREFIX)
	{
		combineList->rep_info[combineList->no_rep_arrays].child_arrays = (void **)malloc(sizeof(void *) * thisgroup->tree->no_children);
		for(i = 0; i < thisgroup->tree->no_children; ++i)
			combineList->rep_info[combineList->no_rep_arrays].child_arrays[i] = (void *)malloc(size);
		combineList->rep_info[combineList->no_rep_arrays].old_base = (void *)malloc(size);
		memcpy(combineList->rep_info[combineList->no_rep_arrays].old_base, _array->base, size);
#ifdef DEBUG
		printf("%d: prefix_array:", NestStep_get_rank());
		for(i = 0; i < 10; ++i)
		{
			if(_array->type == DVAR)
				printf(" %.2f", ((double *)(combineList->rep_info[combineList->no_rep_arrays].old_base))[i]);
		}
		printf("\n");
#endif
	}
	combineList->rep_info[combineList->no_rep_arrays].updated = _updated;
	
	/*
	if(_combine_strategy == UPDATED)
	{
		if(_updated != UPDATED)
			set_RepArray_zero(_array);
		combineList->rep_info[combineList->no_rep_arrays].combine_strategy = ADD;
	}
	*/
	
	combineList->largest_rep_array = (combineList->largest_rep_array < size) ? size : combineList->largest_rep_array;

	++combineList->no_rep_arrays;

	return ATTACH_SUCCESS;
}

/*
 * new_BlockArray
 * creates a new BlockDistArray element and returns a pointer to it
 */
BlockDistArray * new_BlockArray(Name _name, int _type, int _size)
{
	BlockDistArray * temp;
	int element_size;
	int no_elements;
	int ratio;
	int remainder;

	// create the array
	element_size = get_element_size(_type);
	
	ratio = _size / NestStep_get_size();
	remainder = _size % NestStep_get_size();

	no_elements = ratio + ((remainder > NestStep_get_rank()) ? 1 : 0);
	
	temp = (BlockDistArray *)malloc(sizeof(BlockDistArray));
	temp->name.procedure = _name.procedure;
	temp->name.relative = _name.relative;
	memcpy(&(temp->name.path[0]), &(thisgroup->path[0]), MAX_DEPTH * sizeof(int));
	temp->type = _type;
	temp->size = _size;
	temp->local_size = no_elements;
	temp->lower = (NestStep_get_rank() * ratio) + ((remainder > NestStep_get_rank()) ? NestStep_get_rank() : remainder); 
	temp->higher = temp->lower + no_elements - 1;
	temp->base = (void *)malloc(element_size * no_elements);

	NestStep_register_b(arrayHolder, temp);

	return temp;
}

/*
 * free_BlockArray
 * frees the memory of a BlockDistArray
 */
void free_BlockArray(BlockDistArray * _array)
{
	NestStep_unregister_b(arrayHolder, _array);
	free(_array->base);
	free(_array);
}

/*
 * new_CyclicArray
 */
CyclicDistArray * new_CyclicArray(Name _name, int _type, int _size, int _blocksize)
{
	CyclicDistArray * temp;
	int element_size;
	int no_elements;
	int no_blocks;
	int no_blocks_each;
	int no_blocks_left;

	element_size = get_element_size(_type);

	no_blocks_each = _size / (_blocksize * NestStep_get_size());
	no_blocks_left = _size % (_blocksize * NestStep_get_size());
	no_blocks = no_blocks_each + 
		((no_blocks_left > (_blocksize * NestStep_get_rank())) ? 1 : 0);
	no_elements = no_blocks * _blocksize;

	temp = (CyclicDistArray *)malloc(sizeof(CyclicDistArray));
	temp->name.procedure = _name.procedure;
	temp->name.relative = _name.relative;
	memcpy(&(temp->name.path[0]), &(thisgroup->path[0]), MAX_DEPTH * sizeof(int));
	temp->type = _type;
	temp->size = _size;
	temp->blocksize = _blocksize;
	temp->base = (void *)malloc(element_size * no_elements);

	NestStep_register_c(arrayHolder, temp);

	return temp;
}

/*
 * free_CyclicArray
 */
void free_CyclicArray(CyclicDistArray * _array)
{
	NestStep_unregister_c(arrayHolder, _array);
	free(_array->base);
	free(_array);
}

/*
 * compare_array_b
 * compares the names of two blockinfo items
 */
int compare_array_b(const void * _info1, const void * _info2)
{
	const BlockArrayInfo * info1 = (const BlockArrayInfo *)_info1;
	const BlockArrayInfo * info2 = (const BlockArrayInfo *)_info2;

	return compare_name(info1->name, info2->name);
}

/*
 * compare_block_pointers
 */
int compare_block_pointers(const void * _pointer1, const void * _pointer2)
{
	const BlockDistArray ** pointer1 = (const BlockDistArray **)_pointer1;
	const BlockDistArray ** pointer2 = (const BlockDistArray **)_pointer2;

	return compare_name((*pointer1)->name, (*pointer2)->name);
}

/*
 * compare_cyclic_pointers
 */
int compare_cyclic_pointers(const void * _pointer1, const void * _pointer2)
{
	const CyclicDistArray ** pointer1 = (const CyclicDistArray **)_pointer1;
	const CyclicDistArray ** pointer2 = (const CyclicDistArray **)_pointer2;

	return compare_name((*pointer1)->name, (*pointer2)->name);
}

/*
 * free_ArrayHolder
 */
void free_ArrayHolder(ArrayHolder * _holder)
{
	free(_holder->block_arrays);
	free(_holder->cyclic_arrays);
	free(_holder);
}

/*
 * NestStep_register_b
 */
int NestStep_register_b(ArrayHolder * _holder, BlockDistArray * _array)
{
	if(_holder->no_block_arrays < MAX_BLOCK_ARRAYS - 1)
	{
		_holder->block_arrays[_holder->no_block_arrays] = _array;
		++_holder->no_block_arrays;

		qsort(_holder->block_arrays,
			  _holder->no_block_arrays,
			  sizeof(BlockDistArray *),
			  compare_block_pointers);
	}
	else
	{
		return REGISTER_ERROR;
	}

	return REGISTER_SUCCESS;
}

int NestStep_unregister_b(ArrayHolder * _holder, BlockDistArray * _array)
{
	BlockDistArray ** ptr = bsearch(&_array, 
										 _holder->block_arrays, 
										 _holder->no_block_arrays, 
										 sizeof(BlockDistArray *),
										 compare_block_pointers);
	if(ptr != NULL)
	{
		/* remove entry */
		*ptr = _holder->block_arrays[_holder->no_block_arrays - 1];
		_holder->no_block_arrays--;
		qsort(_holder->block_arrays,
			  _holder->no_block_arrays,
			  sizeof(BlockDistArray *),
			  compare_block_pointers);
	}
	return 0;
}

/*
 * NestStep_register_c
 */
int NestStep_register_c(ArrayHolder * _holder, CyclicDistArray * _array)
{
	if(_holder->no_cyclic_arrays < MAX_CYCLIC_ARRAYS - 1)
	{
		_holder->cyclic_arrays[_holder->no_cyclic_arrays] = _array;
		++_holder->no_cyclic_arrays;
		qsort(_holder->cyclic_arrays,
			  _holder->no_cyclic_arrays,
			  sizeof(CyclicDistArray *),
			  compare_cyclic_pointers);
	}
	else
	{
		return REGISTER_ERROR;
	}

	return REGISTER_SUCCESS;
}

int NestStep_unregister_c(ArrayHolder * _holder, CyclicDistArray * _array)
{
	CyclicDistArray ** ptr = bsearch(&_array, 
										 _holder->cyclic_arrays, 
										 _holder->no_cyclic_arrays, 
										 sizeof(CyclicDistArray *),
										 compare_cyclic_pointers);
	if(ptr != NULL)
	{
		/* remove entry */
		*ptr = _holder->cyclic_arrays[_holder->no_cyclic_arrays - 1];
		_holder->no_cyclic_arrays--;
		qsort(_holder->cyclic_arrays,
			  _holder->no_cyclic_arrays,
			  sizeof(CyclicDistArray *),
			  compare_cyclic_pointers);
	}
	return 0;
}

/*
 * locate_array_b
 */
BlockDistArray * locate_array_b(ArrayHolder * _holder, Name _name)
{
	/*BlockDistArray * key = new_BlockArray(_name, IVAR, 0);
	BlockDistArray ** returnptr = bsearch(&key, 
										 _holder->block_arrays, 
										 _holder->no_block_arrays, 
										 sizeof(BlockDistArray *),
										 compare_block_pointers);

	free_BlockArray(key);*/

	/*if(returnptr == NULL)
		return NULL;
	else
		return *(returnptr);*/
	int i;
	for(i = 0; i < _holder->no_block_arrays; i++)
	{
		if(compare_name(_name, _holder->block_arrays[i]->name) == 0)
			return _holder->block_arrays[i];
	}
	return NULL;
}

/*
 * locate_array_c
 */
CyclicDistArray * locate_array_c(ArrayHolder * _holder, Name _name)
{
	/*CyclicDistArray * key = new_CyclicArray(_name, IVAR, NestStep_get_size(), 1);
	CyclicDistArray ** returnptr = bsearch(&key,
										  _holder->cyclic_arrays,
										  _holder->no_cyclic_arrays,
										  sizeof(CyclicDistArray *),
										  compare_cyclic_pointers);
	
	free_CyclicArray(key);

	if(returnptr == NULL)
		return NULL;
	else
		return *(returnptr);*/
	int i;
	for(i = 0; i < _holder->no_cyclic_arrays; i++)
	{
		if(compare_name(_name, _holder->cyclic_arrays[i]->name) == 0)
			return _holder->cyclic_arrays[i];
	}
	return NULL;
}

/*
 * new_ArrayHolder
 */
ArrayHolder * new_ArrayHolder(void)
{
	ArrayHolder * temp = (ArrayHolder *)malloc(sizeof(ArrayHolder));
	temp->block_arrays = (BlockDistArray **)malloc(sizeof(BlockDistArray *) * MAX_BLOCK_ARRAYS);
	temp->no_block_arrays = 0;

	temp->cyclic_arrays = (CyclicDistArray **)malloc(sizeof(CyclicDistArray *) * MAX_CYCLIC_ARRAYS);
	temp->no_cyclic_arrays = 0;

	return temp;
}

/*
 * print_b_arrays_holder
 */
void print_b_arrays_holder(ArrayHolder * _holder)
{
	int i;
	for(i = 0; i < _holder->no_block_arrays; ++i)
	{
		printf("%d:%d: array %d in holder: %d:%d\n",
			NestStep_get_group_id(),
			NestStep_get_rank(),
			i,
			_holder->block_arrays[i]->name.procedure,
			_holder->block_arrays[i]->name.relative);
	}
}

int global_local_index_b(BlockDistArray * _array, int _index)
{
	if(_index < _array->lower || _index > _array->higher)
		return -1;
	else
		return _index - _array->lower;
}

int local_global_index_b(BlockDistArray * _array, int _index)
{
	if(_index < 0 || _index > _array->higher - _array->lower)
		return -1;
	else
		return _index + _array->lower;
}

int global_local_index_c(CyclicDistArray * _array, int _index)
{
	int remainder;
	int ratio;
	
	if(_index < 0 || _index >= _array->size)
		return -1;

	remainder = _index % (_array->blocksize * NestStep_get_size());
	ratio = _index / (_array->blocksize * NestStep_get_size());

	if( remainder >= (NestStep_get_rank() * _array->blocksize) && 
		remainder <  ((NestStep_get_rank() + 1) * _array->blocksize) )
	{
		return ratio * _array->blocksize + remainder - NestStep_get_rank() * _array->blocksize;
	}
	else
	{
		return -1;
	}
}

int local_global_index_c(CyclicDistArray * _array, int _index)
{
	int gindex;
	
	gindex = (_index / _array->blocksize) * _array->blocksize * NestStep_get_size();
	gindex += NestStep_get_rank() * _array->blocksize;
	gindex += _index % _array->blocksize;

	if(gindex < 0 || gindex >= _array->size)
		return -1;
	else
		return gindex;
}

/*
 * NestStep_scatter_rep_to_b
 */
int NestStep_scatter_rep_to_b(RepArray * _reparray, BlockDistArray * _barray)
{
	int element_size;

	if(_reparray->size == _barray->size && _reparray->type == _barray->type)
	{
		element_size = get_element_size(_barray->type);

		memcpy(	_barray->base,
				(char *)_reparray->base + _barray->lower * element_size,
				(_barray->higher - _barray->lower + 1) * element_size );
		return SCATTER_SUCCESS;
	}
	else
		return SCATTER_ERROR;
}

/*
 * NestStep_scatter_rep_to_c
 */
int NestStep_scatter_rep_to_c(RepArray * _reparray, CyclicDistArray * _carray)
{
	int i;
	int start;
	int end;
	int amount;
	int element_size;
	
	if(_reparray->size == _carray->size && _reparray->type == _carray->type)
	{
		element_size = get_element_size(_carray->type);

		for(i = 0;; ++i)
		{
			start = (_carray->blocksize) * NestStep_get_rank() +
				i * (_carray->blocksize) * NestStep_get_size();
			end = start + _carray->blocksize - 1;
			if(start >= _carray->size)
				break;
			if(end >= _carray->size)
				end = _carray->size - 1;
			
			amount = end - start + 1;

			memcpy(	(char *)_carray->base + global_local_index_c(_carray, start) * element_size,
					(char *)_reparray->base + start * element_size,
					amount * element_size);

			if(end >= _carray->size)
				break;
		}

		return SCATTER_SUCCESS;
	}
	else
		return SCATTER_ERROR;
}

/*
 * NestStep_gather_b_to_rep
 */
int NestStep_gather_b_to_rep(BlockDistArray * _barray, RepArray * _reparray)
{
	// CombineList * clist;
	// ArrayHolder * holder;

	if(_barray->size == _reparray->size && _barray->type == _reparray->type)
	{
		// clist = new_CombineList();
		// holder = new_ArrayHolder();
		// NestStep_register_b(holder, _barray);
		
		NestStep_read_request_b(_barray, 0, _barray->size - 1, 
			_reparray->base, NULL); // clist);

		// NestStep_combine(clist, holder);

		// free_CombineList(clist);
		// free_ArrayHolder(holder);

		return GATHER_SUCCESS;
	}
	else
		return GATHER_ERROR;
}

/*
 * NestStep_gather_c_to_rep
 */
int NestStep_gather_c_to_rep(CyclicDistArray * _carray, RepArray * _reparray)
{
	int i;
	int start;
	int end;
	int amount;
	int element_size;
	// CombineList * clist;
	// ArrayHolder * holder;
	
	if(_carray->size == _reparray->size && _carray->type == _reparray->type)
	{
		element_size = get_element_size(_carray->type);

		// clist = new_CombineList();
		// holder = new_ArrayHolder();
		// NestStep_register_c(holder, _carray);
		
		NestStep_read_request_c(_carray, 0, _carray->size - 1, 
			_reparray->base, NULL); // clist);

		// NestStep_combine(clist, holder);

		// free_CombineList(clist);
		// free_ArrayHolder(holder);

		return GATHER_SUCCESS;
	}
	else
		return GATHER_ERROR;
}

//int randomLocalIndex_b(BlockDistArray * _array)
//{
//	return rand() % (_array->higher - _array->lower + 1);
//}
//
//int randomLocalIndex_c(CyclicDistArray * _array)
//{
//	int global;
//	int local;
//
//	do
//	{
//		global = rand() % _array->size;
//	} while((local = global_local_index_c(_array, global)) == -1);
//
//	return local;
//}

void set_RepArray_zero(RepArray * _array)
{
	memset(_array->base, 0, _array->size * get_element_size(_array->type));
}

void set_BlockArray_zero(BlockDistArray * _array)
{
	memset(_array->base, 0, (_array->higher - _array->lower + 1) * get_element_size(_array->type));
}

void set_CyclicArray_zero(CyclicDistArray * _array)
{
	int no_blocks;
	int no_local_blocks;
	int no_elements;

	no_blocks = _array->size / _array->blocksize;
	no_local_blocks = no_blocks / NestStep_get_size();
	no_local_blocks += (no_blocks % NestStep_get_size() > NestStep_get_rank()) ? 1 : 0;
	no_elements = no_local_blocks * _array->blocksize;
	memset(_array->base, 0, no_elements * get_element_size(_array->type));
}


