#ifndef __COMBINE
#define __COMBINE

#include "common.h"
#include "ShVar.h"
#include "arrays.h"
#include <mpi.h>

// macros for messagetypes
#define COMBINE 1
#define COMMIT 2
#define COMB_REP_ARRAY 4
#define COMM_REP_ARRAY 8
#define READ_REQUEST 16
#define WRITE_REQUEST 32
#define REQUEST_ID_BASE 64 // make sure that this is always largest

// maximum array sizes to be sent under UPDATED strategy without
// checking if it is needed to send the array
// the reason for specifying macros for each type instead of a function
// that calculates a size in bytes is because in the case of a homogenous
// cluster alla machines may not represent the types with the same amount
// of bytes
// a function (array_size) is provided for determining whether the array is to big,
// these macros should only be used in that function
// array_size returns SMALL, LARGE or ARRAY_SIZE_ERROR
#define MAX_REP_ARRAY_INT 64
#define MAX_REP_ARRAY_FLOAT 64
#define MAX_REP_ARRAY_DOUBLE 32
#define SMALL 0
#define LARGE 1
#define ARRAY_SIZE_ERROR 2

typedef union value {
	int i;		/* IVAR  */
	float f;	/* FVAR  */
	double d;	/* DVAR  */
} Value;

typedef struct combineitem {
	Name name;				// for returning prefix variables this is the name of the shared var
	union value val;
	int contributor;		/* process who the combineitem originates from */
	int prefix_tag;			// NO_PREFIX, PREFIX or UPDATE, used to differ between updated and not updated items too
} CombineItem;				// UPDATE is only for combineitems used to send prefixvalues back

typedef struct combineinfo {
	ShVar * var;
	int type;						// CHVAR, IVAR, FVAR, DVAR
	int combine_strategy;			// ADD, MULT, AND, OR, FUNC, LEADER, UPDATED
	void (*combineFn)(void *, void *, void *);	// combine function
	int updated;			// NOT_UPDATED, UPDATED
} CombineInfo;

typedef struct prefixvar {
	void * prefix;
	struct shvar * shared;
	Value from_parent;
	Value my_value;
	Value * child_values;
} PrefixVar;

typedef struct reparrayinfo {
	RepArray * rep_array;
	int type;
	int combine_strategy;
	void (*combineFn)(void *, void *, void *);
	int prefix;
	void * old_base;
	void * prefix_base;
	void ** child_arrays;
	int updated;
} RepArrayInfo;

typedef struct combinelist {
	CombineItem * combine_items;
	CombineInfo * combine_info;
	int no_combine_items;
	PrefixVar * prefix_vars;
	int no_prefix_vars;
	RepArrayInfo * rep_info;
	int no_rep_arrays;
	int largest_rep_array;
	BlockArrayInfo * block_info;
	int no_block_arrays;
	CyclicArrayInfo * cyclic_info;
	int no_cyclic_arrays;
	int request_id_counter;
	int * N_DAmsgs;
} CombineList;

/* internal functions */
int combine_send_ch(CombineList * _clist);
int combine_send_p(CombineList * _clist);
int combine_receive_ch(CombineList * _clist);
int combine_receive_p(CombineList * _clist);
int clear_CombineList(CombineList * _clist);

CombineInfo * lookup_info(CombineItem item, CombineList * _clist);
PrefixVar * lookup_prefix(CombineItem item, CombineList * _clist);
int compare_info(const void * _info1, const void * _info2);
int compare_prefix(const void * _prefix1, const void * _prefix2);
int merge(CombineInfo * _info, int _i, CombineList * _clist);
int merge_values(Value * _dest, Value _val1, Value _val2, int _type,
				 int _combine_strategy, void (*combineFn)(void *, void *, void *));
int merge_ints(Value * _dest, Value _val1, Value _val2, int _combine_strategy,
			   void (*combineFn)(void *, void *, void *));
int merge_floats(Value * _dest, Value _val1, Value _val2, int _combine_strategy,
				 void (*combineFn)(void *, void *, void *));
int merge_doubles(Value * _dest, Value _val1, Value _val2, int _combine_strategy,
				  void (*combineFn)(void *, void *, void *));
Value to_value(void * _base, int _type);
int to_real_value(void * _base, int _type, Value _val);

CombineInfo * new_CombineInfo(ShVar * var,
							  int combine_strategy, 
							  void (*combineFn)(void *, void *, void *));

void free_CombineInfo(CombineInfo * info);
int merge_real_i(int * _base, int _one, int _two, int _combine_strategy,
				 void (*combineFn)(void *, void *, void *));
int merge_real_f(float * _base, float _one, float _two, int _combine_strategy,
				 void (*combineFn)(void *, void *, void *));
int merge_real_d(double * _base, double _one, double _two, int _combine_strategy,
				 void (*combineFn)(void *, void *, void *));

int merge_rep_arrays(void * _dest, void * _buffer, int _type, int _size,
				 int _combine_strategy,	void (*_combineFn)(void *, void *, void *), void * hidden);
int merge_rep_arrays_i(int * _dest, int * _buffer, int _size, int _combine_strategy,
				   void (*_combineFn)(void *, void *, void *), int * hidden);
int merge_rep_arrays_f(float * _dest, float * _buffer, int _size, int _combine_strategy,
				   void (*_combineFn)(void *, void *, void *), float * hidden);
int merge_rep_arrays_d(double * _dest, double * _buffer, int _size, int _combine_strategy,
				   void (*_combineFn)(void *, void *, void *), double * hidden);

int array_size(int _elements, int _type);

int process_DA_requests(CombineList * _clist, ArrayHolder * _holder);
int process_DA_incoming(CombineList * _clist, ArrayHolder * _holder);
int process_DA_read(ArrayRequest _request);
int process_DA_write(ArrayRequest _request);

MPI_Datatype get_MPI_type(int _type);
int get_element_size(int _type);

extern CombineItem * srbuffer;

#endif
