#ifndef __NESTSTEP
#define __NESTSTEP

#include "tlib.h"
#include "common.h"
#include "ShVar.h"
#include "Combine.h"
#include "arrays.h"

typedef struct tree {
	int parent;
	int no_children;
	int * children;
} Tree;

typedef struct group {
	T_Descr t_group;
	struct group * parent;
	struct tree * tree;
	int path[MAX_DEPTH];
	CombineList * clist;
} Group;

extern Group * root;
extern Group * thisgroup;
extern MPI_Datatype combine_item_type;
extern MPI_Datatype array_request_type;
extern BlockDistArray * block_arrays;
extern CyclicDistArray * cyclic_arrays;
extern void * bbuffer;
extern ArrayHolder * arrayHolder;

/* in NestStep.c */

// needs to be run before any other NestStep function
void NestStep_init(int * pargc, char *** pargv);

// last NestStep function to be run
void NestStep_finalize(void);

// retreives the rank of the processor in the current group
int  NestStep_get_rank(void);

// retreives the group size of the current group
int	 NestStep_get_size(void);

// get the group number
int  NestStep_get_group_id(void);

// return a pointer to the parent group
Group * NestStep_get_parent_group(void);

// NestStep_step should be called when a step statement occurs
void NestStep_step(void);

// ends current step
void NestStep_step_end(void);

// NestStep_neststep(int k) is a nested step that creates k subgroups
void NestStep_neststep(int k);

// NestStep_neststep_w(int k, float weight[]) is a nested step that creates
// k subgroups, with groupsizes that is specified by the weight array
void NestStep_neststep_w(int k, float weight[]);

// NestStep_neststep_c(int k, int color) is a nested step where color
// decides which subgroup the processor ends up in
void NestStep_neststep_c(int k, int color);

// NestStep_neststep_end ends a nested step
void NestStep_neststep_end(void);

// ignore compare_name, should reside somewhere else
int compare_name(Name name1, Name name2);
// ignore new_Group and free_Group, should reside somewhere else
Group * new_Group(void);
void free_Group(Group * group);
// ignore new_Tree and free_Tree, should reside somewhere else
int new_Tree(void);
int free_Tree(void);

/* in ShVar.c */

// new_ShVar
// creates a shared variable, type is IVAR, FVAR or DVAR (int, float, double)
// base is the address of the "real" variable
ShVar * new_ShVar(Name name, int type, void * base);
void free_ShVar(ShVar * var);

// NestStep_shvar_attach
// attaches a shared variable to a combinelist that will be used in the combinephase
// var = the shared variable, combine_strategy = method for combining the elements,
// combineFn is NULL or a function that combines elements
// (arg1, arg2 pointers to operands, arg3 pointer to where the result should be stored)
// prefix_tag = PREFIX or NO_PREFIX, prefix_base = NULL or address to where the prefix should be stored
// _updated = UPDATED or NOT_UPDATED, _clist = combinelist that the variable should be added to
int NestStep_shvar_attach(ShVar * var, int combine_strategy,
						  void (*combineFn)(void *, void *, void *),
						  int prefix_tag, void * prefix_base,
						  int _updated, CombineList * _clist);

/* in arrays.c */

// new_RepArray
// _name = a unique name for the array, type = IVAR, FVAR or DVAR,
// size = number of elements in the array, base = pointer to base of 
// "real" array
RepArray * new_RepArray(Name _name, int _type, int _size, void * _base);
void free_RepArray(RepArray * _array);

// set_RepArray_zero
// sets all elements in the array to zero
void set_RepArray_zero(RepArray * _array);
// set_BlockArray_zero
// clear the local elements
void set_BlockArray_zero(BlockDistArray * _array);
// set_CyclicArray_zero
// clear the local elements
void set_CyclicArray_zero(CyclicDistArray * _array);

// These functions transform from global to local indexes and
// the other way around. If the input index is invalid -1 is
// returned
int global_local_index_b(BlockDistArray * _array, int _index);
int local_global_index_b(BlockDistArray * _array, int _index);
int global_local_index_c(CyclicDistArray * _array, int _index);
int local_global_index_c(CyclicDistArray * _array, int _index);


// NestStep_RepArray_range
// set the range for which elements to update
// the range is set until after the next combine
int NestStep_RepArray_range(RepArray * _array, int _lower, int _higher);

// NestStep_RepArray_attach
// attaches an array to a combinelist, see NestStep_shvar_attach
int NestStep_RepArray_attach(RepArray * _array,
							 int _combine_strategy,
							 void (*_combineFn)(void *, void *, void *),
							 int _prefix_tag,
							 void * _prefix_base,
							 int _updated, 
							 CombineList * _clist);

// new_BlockArray
// creates a new block array, _size = size of all elements, not individual blocks
// does not take a pointer to an existing array, creates space on the heap
BlockDistArray * new_BlockArray(Name _name, int _type, int _size);
void free_BlockArray(BlockDistArray * _array);

// new_CyclicArray
// creates a new cyclic array, _size = size of all elements,
// _blocksize = size of the blocks that should be distributed in a cyclic manner
// _size / _blocksize must not leave a remainder
CyclicDistArray * new_CyclicArray(Name _name, int _type, int _size, int _blocksize);
void free_CyclicArray(CyclicDistArray * _array);

// new_ArrayHolder
// creates a holder to store information about distributed arrays so
// that they can be found for reads/writes to non-local elements
ArrayHolder * new_ArrayHolder(void);
void free_ArrayHolder(ArrayHolder * _holder);

// following functions are for registering arrays with an ArrayHolder
int NestStep_register_b(ArrayHolder * _holder, BlockDistArray * _array);
int NestStep_register_c(ArrayHolder * _holder, CyclicDistArray * _array);
int NestStep_unregister_b(ArrayHolder * _holder, BlockDistArray * _array);
int NestStep_unregister_c(ArrayHolder * _holder, CyclicDistArray * _array);

// NestStep_scatter_rep_to_b
// scatters the elements of an replicated array over a block array of
// equal size. It is assumed that the replicated array is equal for all
// nodes
int NestStep_scatter_rep_to_b(RepArray * _reparray, BlockDistArray * _barray);
// NestStep_scatter_rep_to_c
// same as above except that it is for cyclic arrays
int NestStep_scatter_rep_to_c(RepArray * _reparray, CyclicDistArray * _carray);
// NestStep_gather_b_to_rep
// gathers the elements of a block array into an replicated one
int NestStep_gather_b_to_rep(BlockDistArray * _barray, RepArray * _reparray);
// NestStep_gather_c_to_rep
// gathers the elements of a cyclic array into an replicated one
int NestStep_gather_c_to_rep(CyclicDistArray * _carray, RepArray * _reparray);

/* in Combine.c */

// new_CombineList
// creates a CombineList
CombineList * new_CombineList(void);
void free_CombineList(CombineList * _clist);

// NestStep_combine
// combines the elements that has been attached to the combinelist provided
// for distributed reads/writes it searches the ArrayHolder for arrays
// NOTE: Needs to be run before NestStep_step_end/NestStep_neststep_end
int NestStep_combine(CombineList * _clist, ArrayHolder * _holder);

// NestStep_read_request_?
// reads elements from the global indexes _lower to _higher into _buffer
int NestStep_read_request_b(BlockDistArray * _array, int _lower, int _higher, 
							void * _buffer, CombineList * _clist);
int NestStep_read_request_c(CyclicDistArray * _array, int _lower, int _higher,
							void * _buffer, CombineList * _clist);
// NestStep_write_request_?
// writes elements to the global indexes _lower to _higher from _buffer
int NestStep_write_request_b(BlockDistArray * _array, int _lower, int _higher, 
							 void * _buffer, CombineList * _clist);
int NestStep_write_request_c(CyclicDistArray * _array, int _lower, int _higher,
							 void * _buffer, CombineList * _clist);

/* in import_export.c */
// NOTE: these functions needs to be executed by all nodes in
// the subgroups. The arrays that is imported from or exported to
// needs to be available in the parent group.
// They should be run after a NestStep_neststep* function but
// before the corresponding NestStep_neststep_end function.

// importArray_rep_to_rep
// imports a subset of the _from array to a the smaller _to array
// _higher - _lower + 1 should equal the smaller ones size
int importArray_rep_to_rep(RepArray * _from, RepArray * _to, 
						   int _lower, int _higher, int active);
// exportArray_rep_to_rep
// exports the smaller _from array to the larger _to array
// _higher - _lower + 1 should equal the smaller ones size
int exportArray_rep_to_rep(RepArray * _from, RepArray * _to,
						   int _lower, int _higher, int active);

// importArray_b_to_b
// imports a the blockarray of the parent group to an equally sized
// blockarray of the current group
int importArray_b_to_b(BlockDistArray * _from, BlockDistArray * _to,
					   int _lower, int _higher, int active);
// exportArray_b_to_b
// exports a the blockarray of the parent group to an equally sized
// blockarray of the current group
int exportArray_b_to_b(BlockDistArray * _from, BlockDistArray * _to,
					   int _lower, int _higher, int active);
// importArray_b_to_rep
// imports a part of a replicated array into a a replicated array
int importArray_b_to_rep(BlockDistArray * _barray, RepArray * _rarray,
						 int _lower, int _higher, int active);
// exportArray_rep_to_b
// imports a replicated array from parent group to a blockarray 
int exportArray_rep_to_b(RepArray * _rarray, BlockDistArray * _barray,
						 int _lower, int _higher, int active);

// importArray_c_to_rep
// see above
int importArray_c_to_rep(CyclicDistArray * _carray, RepArray * _rarray,
						 int _lower, int _higher, int active);
// exportArray_rep_to_c
// see above
int exportArray_rep_to_c(RepArray * _rarray, CyclicDistArray * _carray,
						 int _lower, int _higher, int active);

/* in tree_builders.c */
// changes the tree structure to a D-ary tree
int change_tree_dary(int _dval);

// changes the tree structure to a flat tree
int change_tree_flat(void);

// changes the tree structure to a binomial tree (fraction alpha == bval)
int change_tree_bin(double _bval);

// rebuild the tree(i.e. commit changes)
int rebuild_tree(void);


#endif
