#ifndef TLIB_H
#define TLIB_H

#include <mpi.h>

typedef enum { VALID, ERROR, INVALID } Tstate;

#define ERROR_NONE            0
#define ERROR_ILLEGAL_SPLIT -10
#define ERROR_INVALID_ARGS  -11
#define ERROR_INTERNAL_NULL -12

/** 
 * structure describes the T-group-structures 
 * within each singleprocessor 
 */
typedef struct T_Descr
{ 
  Tstate status;     /**< indicates whether T_Descr exists             */
  char * name;       /**< character string as name of the group        */
  int size;          /**< size of my  group                            */
  int my_rank;       /**< processor id in root group                   */
  int my_rank_local; /**< processor id in current leaf group           */
  int first_rank;    /**< rank of the 1st proc in the corresp. group   */
  int *my_group;     /**< processor ids of all proc. in the group      */
  int my_group_nr;   /**< number of the processor group containing me  */
  int no_of_groups;  /**< number of sibling groups at same level       */
  int **all_groups;  /**< processor ids of all proc. in sibling groups */
  int *group_sizes;  /**< sizes of all parallel groups, same level     */
  int *group_leaders;/**< ranks of the 1st processor in each groups    */
  MPI_Comm comm;     /**< pointer to corresponding MPI communicator    */
  MPI_Group mpigroup;/**< pointer to corresponding MPI group           */
  struct T_Descr *parent; /**< pointer to parents group descriptor     */
} T_Descr;


/**
 * initializes the root descriptor 
 */
void T_Init( int *argc, char ***argv, 
	     MPI_Comm comm, T_Descr *pdescr );


/**
 * splits *descr into two groups of size per1 and per2
 * and returns them in *pdescr
 * @see T_Split_GrpParfor
 */
int T_Split_Grp( T_Descr *descr, T_Descr *pdescr, 
		 float per1, float per2);

/**
 * fills a new T descriptor pdescr according to the 
 * specification in 
 * <pre>
 *   no_of_groups = number of new groups 
 *   per = vector with percentage size of new groups 
 * </pre>
 */
int T_Split_GrpParfor(int n_groups, T_Descr *descr,
		      T_Descr *pdescr,
		      float *per);

/**
 *  splitting of processor groups with specification according to
 *  MPI_Comm_split 
 * <pre>
 * _color : determines target group
 * _key   : determines position in target group 
 *          if keys are equal the rank decides about the order
 *          (see MPI_Comm_split)
 * </pre>
 */
int T_Split_GrpExpl( T_Descr *descr,
		     T_Descr *pdescr, int color, int key);

/**
 * Executes functions (tasks) f1 and f2 concurrently on a different
 * set of processors. Processors which a descriptor of group 0 call
 * function f1 and the processors of group 1 call f2. Not
 * participating processors wait until the two processor groups have
 * completed their computation.
 * @see T_Parfor
 */
int T_Par(void * (*f1)(void *, MPI_Comm, T_Descr *), 
	  void * parg1, 
	  void ** pres1,
	  void * (*f2)(void *, MPI_Comm, T_Descr *),
	  void * parg2, 
	  void ** pres2,
	  T_Descr *descr);

/**
 * Processor group with number <i>i</i> executes function f[<i>i</i>]
 * in parallel. The group number can be determined by
 * <pre>
 *   int T_GetMyGroupNo (T_Descr *descr)
 * </pre>
 * @see T_Par
 */
int T_Parfor(void * (*f[])(void *, MPI_Comm, T_Descr *),
             void * parg[],
             void * pres[],
             T_Descr *descr);

/**
 * returns size of the group that the calling proc belongs to
 */
int T_GetMyGroupSize (T_Descr *descr);

/**
 * returns the global rank of the calling processor
 */
int T_GetMyRank (T_Descr *descr);

/**
 * returns the local rank of a processor
 */
int T_GetMyRankLocal (T_Descr *descr);

/**
 * allocates and returns an array with the leaders of all sibling
 * groups
 */
int *T_GetGroupLeaders (T_Descr *descr);


/**
 * returns assigned MPI communicator
 */
MPI_Comm T_GetComm (T_Descr *descr);


/**
 * returns the descriptor of parent group
 */
T_Descr * T_GetParent (T_Descr *descr);


/** 
 * returns the number of the corresponding group For each hierachical
 * level group numbering starts at 0.
 */
int T_GetMyGroupNo (T_Descr *descr);

/**
 * returns ids of current group
 */
int * T_GetMyGroup (T_Descr *descr);


/**
 * transforms a group replicated distribution into a globally        
 * replicated distribution for two groups                            
 * source1 is available replicated on processor group 1              
 * source2 is available replicated on processor group 2              
 */
int T_ReDistrRepl (void * source1, void * target1, int size1,
		   void * source2, void * target2, int size2,
		   T_Descr *parent, T_Descr *child);

/**
 * sink of the computation
 */
void T_Finalize();  /* added by S.H. */


/**
 * copy a complete T descriptor structure
 */
void copy_T_Descr (T_Descr *source, T_Descr *target);

/**
 * prints the descriptor structure 
 */
void print_T_Descr(T_Descr *pdescr);

/**
 * returns the size of each subgroup when splitting a
 * group with <i>p</i> processors into <i>subg_num</i>
 * subgroups
 */
int *split_group(int p, float *per, int subg_num);

#endif
